/*-
 * Copyright (C)2024 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2024 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: pgoblin0file.c,v 1.9 2025/02/24 20:54:23 babolo Exp $"

#define BLIN_COMPAT      4
#define Bpars_COMPAT     4
#define MULAR_COMPAT     0
#define MIFE_COMPAT      5
#define PGOBLIN_COMPAT   5
#define PGOBLIN_INTERNAL 1
#define PGOBLIN_MODULE   pgoblin_module

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sysexits.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <multilar.h>
#include <mife.h>
#include "pgoblin.h"
#include "options.h"

#define BLIN_5STO24G ((blin_statectl){{5, 1, 8, 24, ' ', 5, 1, 1, 0}})

#define Cm 0x800000  /*  UGO        */
#define CU 0x400000  /*  U        */
#define CG 0x200000  /*  G        */
#define CO 0x100000  /*  O        */
#define Cu 0x080000  /*  u                  */
#define Cg 0x040000  /*  g                  */
#define Co 0x020000  /*  o                  */
#define Cr 0x010000  /*  r                  */
#define Cs 0x008000  /*  s                  */
#define Cw 0x004000  /*  w                  */
#define Cx 0x002000  /*  x                  */
#define Ce 0x001000  /* =                   */
#define Cp 0x000800  /* +                   */
#define Cz 0x000200  /*    */
#define C_state 0x0F /*   */

struct pgoblin_realiod {
          FILE *fout;
          int   onu;
          int   openflags;
};

typedef enum classes
{ L0             /* \0     */
, Lc             /* ,      */
, La, Lu, Lg, Lo /* who    */
, Lp, Le         /* op     */
, Lr, Ls, Lw, Lx /* perm   */
, Lz             /* other  */
, Ly             /* border */
} classes;

typedef enum states
{ bw /* who     */
, be /* after = */
, bp /* after + */
, bz /* end     */
} states;

extern pgoblin_ios PGOBLIN_MODULE;

static const char   statenames[][3] = { "bw", "be", "bp", "bz"};

static const char  *clasnames[] =
{ "L0", "Lc", "La", "Lu", "Lg", "Lo", "Lp", "Le"
, "Lr", "Ls", "Lw", "Lx", "Lz", "Ly"
};

static const char pn[][3] = 
{ "- ", "- ", "- ", "- ", "- ", "- ", "- ", "- "
, "m", "CU", "CG", "CO", "Cu", "Cg", "Co", "Cr"
, "Cs", "Cw", "Cx", "Ce", "Cp", "- ", "Cz", "- "
};

static const u_char class[256] =
{  L0, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz
,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz
,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lp,  Lc, Lz, Lz, Lz
,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Le, Lz, Lz

,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz
,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz
,  Lz, La, Lz, Lz,  Lz, Lz, Lz, Lg,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lo
,  Lz, Lz, Lr, Ls,  Lz, Lu, Lz, Lw,  Lx, Lz, Lz, Lz,  Lz, Lz, Lz, Lz

,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz
,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz
,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz
,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz

,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz
,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz
,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz
,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz,  Lz, Lz, Lz, Lz
};

static u_int32_t automa[Ly][bz] =
/*           bw           be        bp    */
{ {       Cz|bz, Cm|Ce|Cp|bz, Cm|Cp|bz } /* L0 0 */
, {       Cz|bz, Cm|Ce|Cp|bw, Cm|Cp|bw } /* Lc , */
, { CU|CG|CO|bw,       Cz|bz,    Cz|bz } /* La   */
, {       CU|bw,       Cu|be,    Cu|bp } /* Lu   */
, {       CG|bw,       Cg|be,    Cg|bp } /* Lg   */
, {       CO|bw,       Co|be,    Co|bp } /* Lo   */
, {          bp,    Ce|Cp|bp,    Cp|bp } /* Lp + */
, {          be,    Ce|Cp|be,    Cp|be } /* Le = */
, {       Cz|bz,       Cr|be,    Cr|bp } /* Lr   */
, {       Cz|bz,       Cs|be,    Cs|bp } /* Ls   */
, {       Cz|bz,       Cw|be,    Cw|bp } /* Lw   */
, {       Cz|bz,       Cx|be,    Cx|bp } /* Lx   */
, {       Cz|bz,       Cz|bz,    Cz|bz } /* Lz   */
};

static int
/*****************************************************************************************************
 **                                                                                                 **/
mode(char c, const char *o, mode_t old) {                                                          /**
 **                                                                                                 **
 *****************************************************************************************************/
    u_int32_t control;
    states    state;
    mode_t    cook;
    int       own;
    char     *t = NULL;
    int       n = old;
    int       i;

    if  (!(t = malloc(2 + strlen(o)))) {
        ifBLIN_QW0("malloc");
        n = -EX_OSERR;
        goto out;
    }
    *t = c;
    stpcpy(&t[1], o);
    blin_stateheader(BLIN_5STO24G, pn);
    for (state = bw, own = 0, cook = 0, i = 0; state < bz; i++, state = control & C_state) {
        control = automa[class[t[i] & 0x0FF]][state];
        blin_statebody( BLIN_5STO24G
                      , pn
                      , statenames
                      , clasnames[class[*t & 0x0FF]]
                      , t
                      , (size_t)i
                      , control
                      , state
                      , cook
                      , (u_char)own
                      );
        if  (Cm & control) own = 0;
        if  (CU & control) own |= 4;
        if  (CG & control) own |= 2;
        if  (CO & control) own |= 1;
        if  (Cu & control) cook |= ((n & S_IRWXU) >> 6) | ((n & S_ISUID) >> 8);
        if  (Cg & control) cook |= ((n & S_IRWXG) >> 3) | ((n & S_ISGID) >> 7);
        if  (Co & control) cook |= n & S_IRWXO;
        if  (Cr & control) cook |= S_IROTH;
        if  (Cs & control) cook |= 0x08;
        if  (Cw & control) cook |= S_IWOTH;
        if  (Cx & control) cook |= S_IXOTH;
        if  (Ce & control) n = 0;
        if  (Cp & control) {
            if  (!own) own = 7;
            for (int j = 0; !!own; own >>= 1, ++j) {
                if  (1 & own) {
                    n |= (cook << (3 * j));
                    if  (0 < j) n |= S_ISGID << (j - 1);
            }   }
            cook = 0;
        }
        if  (Cz & control) {
            ifBLIN_QX0("Syntax");
            errno = EINVAL;
            n = -EX_USAGE;
            break;
    }   }
out:
    free(t);
    return(n);
}

static int
/*****************************************************************************************************
 **                                                                                                 **/
init(pgoblin_exenv *exenv, u_int slot) {                                                           /**
 **                                                                                                 **
 *****************************************************************************************************/
    (void)exenv;
    (void)slot;
    return(EX_OK);
}

static int
/*****************************************************************************************************
 **                                                                                                 **/
fini(pgoblin_exenv *exenv) {                                                                       /**
 **                                                                                                 **
 *****************************************************************************************************/
    (void)exenv;
    return(EX_OK);
}

static int
/**********************************************************************
 **                                                                  **/
closew(pgoblin_rio *rio) {                                          /**
 **                                                                  **
 **********************************************************************/
#   define blin_internal_flags (rio->flags & BLIN_MASK)
    int   ex = EX_OK;

    ifBLIN_QX3("+ [%d]", rio->flags & PGOBLIN_STYLE_TYPE);
    if  (!!rio && !!rio->iod) {
        MARK_IO_OUT_GO(rio);;
        ;;  if  (!!rio->iod->fout && !!(ex = fdclose(rio->iod->fout, NULL)) && (EFDABSENT != errno)) {
        ;;      ifBLIN_QW0("fdclose fout");
        ;;  } else {
        ;;      ex = EX_OK;
        ;;  }
        ;;  if  (!!rio->freeouts && !!(ex = close(rio->iod->onu))) ifBLIN_QW0("close");
        ;;  free(rio->iod);
        ;;  rio->iod = NULL;
        ;;  rio->parmask &= ~PGOBLIN_OUTSET;
        ;;  rio->freeouts = 0;
        ;;  rio->eof = 0;
        ;;  rio->ioerr = 0;
        ;;  rio->trapdef = 0;
        MARK_IO_OUT_WENT(rio);;
    }
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static int
/**********************************************************************
 **                                                                  **/
oprep(pgoblin_exenv *exenv, pgoblin_nr *rn) {                       /**
 **                                                                  **
 **********************************************************************/
#   define blin_internal_flags (rou->flags & BLIN_MASK)
    int              openflags;
    int              oneflag;
    int              modeflags;
    const char      *flar   = "=:a:f:g:n:o:u:";
    pgoblin_rio     *rou;
    pgoblin_rio     *rin;
    babolo_opts     *bos    = NULL;
    babolo_parm     *bp     = NULL;
    int              ex     = EX_OK;
    int              i;
    const char      *o      = NULL;
    int              c;

    GET_RIO(rou, exenv->options, rn[PGO_COUT]);
    openflags = O_WRONLY;
    modeflags = 0;
    if  (!!rou->path && !!*(rou->path)) {
        bp = babolo_getparm(Bpars_NOAL | Bpars_LFLN | Bpars_NOEN | Bpars_FAST, &rou->path, "#", 0);
        if  (!bp) {
            ifBLIN_QW0("getparm");
            ex = EX_SOFTWARE;
            goto out;
        }
        ex = (int)bp->cntr;
        ifBLIN_QX5("%d %"BLIN_X, bp->argc, BLIN_I(rou->path));
        if  (!(bos = babolo_openopts(Bpars_ESYM | 'E', 'Z'))) {
            ifBLIN_QW0("openopts");
            ex = EX_SOFTWARE;
            goto out;
        }
        if  (!!babolo_setparm(bos, (blin_internal_flags >> 1) & BLIN_MASK, bp, flar)) {
            ifBLIN_QW0("setopts");
            ex = EX_SOFTWARE;
            goto out;
        }
        for (; !!(c = babolo_getopts(bos));) {
            ifBLIN_QX5("-%c", c);
            switch(c) {
            case 'f': if  (!(o = babolo_getoptsarg(bos))) {
                    ;     ifBLIN_QW0("No -%c arg", bos->c);
                    ;     ex = EX_USAGE;
                    ;     goto out;
                    ; }
                    ; if  (0 >= (oneflag = (int)babolo_testword(&options, o))) {
                    ;     ifBLIN_QW0("babolo_testword=%s~", o);
                    ;     ex = EX_SOFTWARE;
                    ;     goto out;
                    ; }
                    ; openflags |= oneflag;
                    ; break;
            case 'n': if  (!(o = babolo_getoptsarg(bos))) {
                    ;     ifBLIN_QW0("No -%c arg", bos->c);
                    ;     ex = EX_USAGE;
                    ;     goto out;
                    ; }
                    ; errno = 0;
                    ; modeflags = (int)strtol(o, NULL, 8);
                    ; if  (!!errno) {
                    ;     ifBLIN_QW0("mode %s", o);
                    ;     ex = EX_USAGE;
                    ;     goto out;
                    ; }
                    ; break;
            case 'a':
            case 'g':
            case 'o':
            case 'u':
                    ; if  (!(o = babolo_getoptsarg(bos))) {
                    ;     ifBLIN_QW0("No -%c arg", bos->c);
                    ;     ex = EX_USAGE;
                    ;     goto out;
                    ; }
                    ; if  (0 > (modeflags = mode((char)c, o, (mode_t)modeflags))) {
                    ;     ifBLIN_QW0("mode %c%s", c, o);
                    ;     ex = EX_USAGE;
                    ;     goto out;
                    ; }
                    ; break;
            case '=': if  (!(o = babolo_getoptsarg(bos))) {
                    ;     ifBLIN_QW0("No -%c arg", bos->c);
                    ;     ex = EX_USAGE;
                    ;     goto out;
                    ; }
                    ; switch(o[0]) {
                      case '0': blin_ctl(BLIN_CTL_DUMP);  exit(EX_OK);
                      case '1': babolo_optest(flar);      exit(EX_OK);
                      case '2': babolo_optext(bos);       exit(EX_OK);
                      case '3': babolo_dumpopts(bos);     exit(EX_OK);
                    ; }
                    ; break;
            default : ifBLIN_QX0("Illegal flag -%c", bos->c);
                    ; ex = EX_USAGE;
                    ; goto out;
        }   }
        o = babolo_getargs(bos);
        ifBLIN_QX4("=%s%s", !o ? "" : o, !o ? "" : "~");
    }
    if  (!!rn[PGO_CIN]) {
        if  (!!o) ifBLIN_QX1("File name overwrite");
        GET_RIO(rin, exenv->options, rn[PGO_CIN]);
        if  (  (0 < pgoblin_db_resinfo(rin, PGOBLIN_Ntuples))
            && (0 < pgoblin_db_resinfo(rin, PGOBLIN_Nfields))
            && !pgoblin_db_valinfo(rin, PGOBLIN_IsNull, 0, 0)
            ) {
            o = pgoblin_db_getvalue(rin, 0, 0);
    }   }
    if  (!o) {
        ifBLIN_QX0("No file name");
        errno = EINVAL;
        ex = EX_USAGE;
        goto out;
    }
    if  (0 > (i = open(o, openflags, modeflags))) {
        ifBLIN_QW0("open=%s~", o);
        ex = EX_IOERR;
        goto out;
    }
    MARK_IO_OUT_GO(rou);;
    ;;  if  (!(rou->iod = calloc(1, sizeof(pgoblin_realiod)))) {
    ;;      ifBLIN_QW0("No mem");
    ;;      ex = EX_OSERR;
    ;;      goto out;
    ;;  }
    ;;  rou->iod->openflags = openflags;
    ;;  rou->iod->onu = i;
    ;;  rou->parmask |= PGOBLIN_OUTSET;
    ;;  rou->freeouts = 1;
    ;;  rou->eof = 0;
    ;;  rou->ioerr = 0;
    ;;  rou->trapdef = 0;
    MARK_IO_OUT_WENT(rou);;
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static ssize_t
/**********************************************************************
 **                                                                  **/
writer(pgoblin_rio *rio, const void *buf, size_t size) {            /**
 **                                                                  **
 **********************************************************************/
#   define blin_internal_flags (rio->flags & BLIN_MASK)
    ssize_t e;

    ifBLIN_QX4( "+ flags %08X e%d t%d %"BLIN_X" %"BLIN_X"=%.*s~"
              , rio->flags
              , rio->eof
              , rio->trapdef
              , BLIN_I(rio)
              , BLIN_I(rio->iod)
              , (int)size, buf
              );
    if  (!!rio && !!rio->iod && !rio->eof) {
        ifBLIN_QX5("%"BLIN_X" %d", BLIN_I(mife_writ), rio->iod->onu);
        if  (0 > (e = mife_writ(rio->iod->onu, buf, size))) {
            ifBLIN_QX6("%"BLIN_D, e);
            if  (!rio->trapdef) {
                ifBLIN_QW0("IO out");
            } else {
                ifBLIN_QW0("IO out %s", rio->ioerr ? "ignore" : "error");
                e = 0;
            }
            if  (!rio->ioerr) {
                rio->ioerr = 1;
            } else {
                rio->eof = 1;
            }
        } else {
            rio->eof = 0;
        }
    } else if (!rio->trapdef) {
        ifBLIN_QX5("");
        e = -1;
    } else {
        ifBLIN_QX5("");
        e = (ssize_t)size;
    }
    ifBLIN_QX4("- %zd", e);
    return(e);
#   undef blin_internal_flags
}

static int
/**********************************************************************
 **                                                                  **/
fuclose(pgoblin_rio *rio) {                                         /**
 **                                                                  **
 **********************************************************************/
#   define blin_internal_flags (rio->flags & BLIN_MASK)
    int ex = EX_OK;

    ifBLIN_QX3("+ [%d]", rio->flags & PGOBLIN_STYLE_TYPE);
    if  (  !!rio
        && !!rio->iod
        && !!rio->iod->fout
        && !!(ex = fdclose(rio->iod->fout, NULL))
        && (EFDABSENT != errno)
        ) {
        ifBLIN_QW1("fdclose");
    } else {
        ex = EX_OK;
    }
    rio->iod->fout = NULL;
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static FILE *
/**********************************************************************
 **                                                                  **/
fuopen(pgoblin_rio *rio) {                                          /**
 **                                                                  **
 **********************************************************************/
#   define blin_internal_flags (rio->flags & BLIN_MASK)
    char f[6];
    int  p;

    ifBLIN_QX3("+ [%d]", rio->flags & PGOBLIN_STYLE_TYPE);
    if  (!!rio && !!rio->iod) {
        if  (!!rio->iod->fout && !!fdclose(rio->iod->fout, NULL) && (EFDABSENT != errno)) {
            ifBLIN_QW1("close at funopen");
        }
        rio->iod->fout = NULL;
        p = 0;
        f[p++] = !(O_APPEND & rio->iod->openflags) ? 'w' : 'a';
        if  (!!(O_EXCL & rio->iod->openflags)) f[p++] = 'x';
        if  (!!(O_CLOEXEC & rio->iod->openflags)) f[p++] = 'e';
        f[p++] = '\0';
        rio->iod->fout = fdopen(rio->iod->onu, f);
        if  (!rio->iod->fout) ifBLIN_QW0("fdopen on %c(%d)", pgoblin_regn[rio->nio], rio->nio);
    }
    ifBLIN_QX3("- %"BLIN_X, BLIN_I(rio->iod->fout));
    return(rio->iod->fout);
#   undef blin_internal_flags
}

static int
/**********************************************************************
 **                                                                  **/
fdout(pgoblin_rio *rio) {                                           /**
 **                                                                  **
 **********************************************************************/
#   define blin_internal_flags (rio->flags & BLIN_MASK)
    int   ex = -1;

    ifBLIN_QX3("+ [%d]", rio->flags & PGOBLIN_STYLE_TYPE);
    if  (!!rio && !!rio->iod) ex = rio->iod->onu;
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

pgoblin_ios PGOBLIN_MODULE =
{ 0                /* flags */
, "0pGoblin-" VERS /* id    */
, "file\0" VERS    /* name  */
, 0                /* dummy */
, init
, fini
, NULL             /* iprep */
, oprep
, NULL             /* read  */
, writer
, NULL             /* seek  */
, fuopen
, fuclose
, NULL             /* fdin  */
, fdout
, closew
};
